import React, { useCallback, useContext, useEffect, useState } from 'react';
import { useConnection } from './connection';
import { useWallet } from './wallet';
import { AccountInfo, Connection, PublicKey } from '@solana/web3.js';
import { AccountLayout, MintInfo, MintLayout, u64 } from '@solana/spl-token';
import { PoolInfo, SwapTokenAccount } from './swapTypes';
import { usePools, useSwapContext } from './swap';
import { WRAPPED_SOL_MINT } from '@project-serum/serum/lib/token-instructions';
import { notify } from './notifications';

const AccountsContext = React.createContext<any>(null);

class AccountUpdateEvent extends Event {
  static type = 'AccountUpdate';
  id: string;
  constructor(id: string) {
    super(AccountUpdateEvent.type);
    this.id = id;
  }
}

class EventEmitter extends EventTarget {
  raiseAccountUpdated(id: string) {
    this.dispatchEvent(new AccountUpdateEvent(id));
  }
}

const accountEmitter = new EventEmitter();

const mintCache = new Map<string, Promise<MintInfo>>();
const pendingAccountCalls = new Map<string, Promise<SwapTokenAccount>>();
const accountsCache = new Map<string, SwapTokenAccount>();

const getAccountInfo = async (connection: Connection, pubKey: PublicKey) => {
  const info = await connection.getAccountInfo(pubKey);
  if (info === null) {
    throw new Error('Failed to find mint account');
  }

  const buffer = Buffer.from(info.data);

  const data = deserializeAccount(buffer);

  const details = {
    pubkey: pubKey,
    account: {
      ...info,
    },
    info: data,
  } as SwapTokenAccount;

  return details;
};

const getMintInfo = async (connection: Connection, pubKey: PublicKey) => {
  const info = await connection.getAccountInfo(pubKey);
  if (info === null) {
    throw new Error('Failed to find mint account');
  }

  const data = Buffer.from(info.data);

  return deserializeMint(data);
};

export const cache = {
  getAccount: async (connection: Connection, pubKey: string | PublicKey) => {
    let id: PublicKey;
    if (typeof pubKey === 'string') {
      id = new PublicKey(pubKey);
    } else {
      id = pubKey;
    }

    const address = id.toBase58();

    let account = accountsCache.get(address);
    if (account) {
      return account;
    }

    let query = pendingAccountCalls.get(address);
    if (query) {
      return query;
    }

    query = getAccountInfo(connection, id).then((data) => {
      pendingAccountCalls.delete(address);
      accountsCache.set(address, data);
      return data;
    }) as Promise<SwapTokenAccount>;
    pendingAccountCalls.set(address, query as any);

    return query;
  },
  getMint: async (connection: Connection, pubKey: string | PublicKey) => {
    let id: PublicKey;
    if (typeof pubKey === 'string') {
      id = new PublicKey(pubKey);
    } else {
      id = pubKey;
    }

    let mint = mintCache.get(id.toBase58());
    if (mint) {
      return mint;
    }

    let query = await getMintInfo(connection, id);

    mintCache.set(id.toBase58(), query as any);

    return query;
  },
};

export const getCachedAccount = (
  predicate: (account: SwapTokenAccount) => boolean,
) => {
  for (const account of accountsCache.values()) {
    if (predicate(account)) {
      return account as SwapTokenAccount;
    }
  }
};

function wrapNativeAccount(
  pubkey: PublicKey,
  account?: AccountInfo<Buffer>,
): SwapTokenAccount | undefined {
  if (!account) {
    return undefined;
  }

  return {
    pubkey: pubkey,
    account,
    info: {
      mint: WRAPPED_SOL_MINT,
      owner: pubkey,
      amount: new u64(account.lamports),
      delegate: null,
      delegatedAmount: new u64(0),
      isInitialized: true,
      isFrozen: false,
      isNative: true,
      rentExemptReserve: null,
      closeAuthority: null,
    },
  };
}

export function UserAccountsProvider({ children = null as any }) {
  const connection = useConnection();
  const { wallet, connected } = useWallet();
  const [tokenAccounts, setSwapTokenAccounts] = useState<SwapTokenAccount[]>(
    [],
  );
  const [userAccounts, setUserAccounts] = useState<SwapTokenAccount[]>([]);
  const [nativeAccount, setNativeAccount] = useState<AccountInfo<Buffer>>();
  const { pools } = usePools();
  const { tokenProgramId } = useSwapContext();

  const selectUserAccounts = useCallback(() => {
    return [...accountsCache.values()].filter(
      (a) => a.info.owner.toBase58() === wallet.publicKey.toBase58(),
    );
  }, [wallet]);

  useEffect(() => {
    setUserAccounts(
      [
        wrapNativeAccount(wallet.publicKey, nativeAccount),
        ...tokenAccounts,
      ].filter((a) => a !== undefined) as SwapTokenAccount[],
    );
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [nativeAccount, tokenAccounts]);

  useEffect(() => {
    if (!connection || !wallet || !wallet.publicKey) {
      setSwapTokenAccounts([]);
    } else {
      const querySwapTokenAccounts = async () => {
        // user accounts are update via ws subscription
        const accounts = await connection.getTokenAccountsByOwner(
          wallet.publicKey,
          {
            programId: tokenProgramId,
          },
        );
        accounts.value
          .map((info) => {
            const data = deserializeAccount(info.account.data);
            // need to query for mint to get decimals

            // TODO: move to web3.js for decoding on the client side... maybe with callback
            const details = {
              pubkey: info.pubkey,
              account: {
                ...info.account,
              },
              info: data,
            } as SwapTokenAccount;

            return details;
          })
          .forEach((acc) => {
            accountsCache.set(acc.pubkey.toBase58(), acc);
          });
        setSwapTokenAccounts(selectUserAccounts());
      };

      querySwapTokenAccounts();

      connection.getAccountInfo(wallet.publicKey).then((acc) => {
        if (acc) {
          setNativeAccount(acc);
        }
      });
      connection.onAccountChange(wallet.publicKey, (acc) => {
        if (acc) {
          setNativeAccount(acc);
        }
      });

      // This can return different types of accounts: token-account, mint, multisig
      // TODO: web3.js expose ability to filter. discuss filter syntax
      const tokenSubID = connection.onProgramAccountChange(
        tokenProgramId,
        (info) => {
          // TODO: fix type in web3.js
          const id = (info.accountId as unknown) as string;
          // TODO: do we need a better way to identify layout (maybe a enum identifing type?)
          if (info.accountInfo.data.length === AccountLayout.span) {
            const data = deserializeAccount(info.accountInfo.data);
            // TODO: move to web3.js for decoding on the client side... maybe with callback
            const details = {
              pubkey: new PublicKey((info.accountId as unknown) as string),
              account: {
                ...info.accountInfo,
              },
              info: data,
            } as SwapTokenAccount;

            if (
              details.info.owner.toBase58() === wallet?.publicKey?.toBase58() ||
              accountsCache.has(id)
            ) {
              accountsCache.set(id, details);
              setSwapTokenAccounts(selectUserAccounts());
              accountEmitter.raiseAccountUpdated(id);
            }
          } else if (info.accountInfo.data.length === MintLayout.span) {
            // TODO: update mints that are already in the cache

            if (mintCache.has(id)) {
              const data = Buffer.from(info.accountInfo.data);
              const mint = deserializeMint(data);
              mintCache.set(id, new Promise((resolve) => resolve(mint)));
              accountEmitter.raiseAccountUpdated(id);
            }

            accountEmitter.raiseAccountUpdated(id);
          }
        },
        'singleGossip',
      );

      return () => {
        connection.removeProgramAccountChangeListener(tokenSubID);
      };
    }
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [connected, wallet?.publicKey]);

  return (
    <AccountsContext.Provider
      value={{
        userAccounts,
        pools,
        nativeAccount,
      }}
    >
      {children}
    </AccountsContext.Provider>
  );
}

export function useNativeAccount() {
  const context = useContext(AccountsContext);
  return {
    account: context.nativeAccount as AccountInfo<Buffer>,
  };
}

export function useMint(id?: string) {
  const connection = useConnection();
  const [mint, setMint] = useState<MintInfo>();

  useEffect(() => {
    if (!id) {
      return;
    }

    cache
      .getMint(connection, id)
      .then(setMint)
      .catch((err) =>
        notify({
          message: err.message,
          type: 'error',
        }),
      );
    const onAccountEvent = (e: Event) => {
      const event = e as AccountUpdateEvent;
      if (event.id === id) {
        cache.getMint(connection, id).then(setMint);
      }
    };

    accountEmitter.addEventListener(AccountUpdateEvent.type, onAccountEvent);
    return () => {
      accountEmitter.removeEventListener(
        AccountUpdateEvent.type,
        onAccountEvent,
      );
    };
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [id]);

  return mint;
}

export function useUserAccounts() {
  const context = useContext(AccountsContext);
  return {
    userAccounts: context.userAccounts as SwapTokenAccount[],
  };
}

export function useAccount(pubKey?: PublicKey) {
  const connection = useConnection();
  const [account, setAccount] = useState<SwapTokenAccount>();

  useEffect(() => {
    const query = async () => {
      try {
        if (!pubKey) {
          return;
        }

        const acc = await cache.getAccount(connection, pubKey).catch((err) =>
          notify({
            message: err.message,
            type: 'error',
          }),
        );
        if (acc) {
          setAccount(acc);
        }
      } catch (err) {
        console.error(err);
      }
    };

    query();

    const onAccountEvent = (e: Event) => {
      const event = e as AccountUpdateEvent;
      if (event.id === pubKey?.toBase58()) {
        query();
      }
    };

    accountEmitter.addEventListener(AccountUpdateEvent.type, onAccountEvent);
    return () => {
      accountEmitter.removeEventListener(
        AccountUpdateEvent.type,
        onAccountEvent,
      );
    };
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [connection, pubKey?.toBase58()]);

  return account;
}

export function useCachedPool() {
  const context = useContext(AccountsContext);
  return {
    pools: context.pools as PoolInfo[],
  };
}

export const useSelectedAccount = (account: string) => {
  const { userAccounts } = useUserAccounts();
  const index = userAccounts.findIndex(
    (acc) => acc.pubkey.toBase58() === account,
  );

  if (index !== -1) {
    return userAccounts[index];
  }

  return;
};

export const useAccountByMint = (mint: string) => {
  const { userAccounts } = useUserAccounts();
  const index = userAccounts.findIndex(
    (acc) => acc.info.mint.toBase58() === mint,
  );

  if (index !== -1) {
    return userAccounts[index];
  }

  return;
};

// TODO: expose in spl package
const deserializeAccount = (data: Buffer) => {
  const accountInfo = AccountLayout.decode(data);
  accountInfo.mint = new PublicKey(accountInfo.mint);
  accountInfo.owner = new PublicKey(accountInfo.owner);
  accountInfo.amount = u64.fromBuffer(accountInfo.amount);

  if (accountInfo.delegateOption === 0) {
    accountInfo.delegate = null;
    accountInfo.delegatedAmount = new u64(0);
  } else {
    accountInfo.delegate = new PublicKey(accountInfo.delegate);
    accountInfo.delegatedAmount = u64.fromBuffer(accountInfo.delegatedAmount);
  }

  accountInfo.isInitialized = accountInfo.state !== 0;
  accountInfo.isFrozen = accountInfo.state === 2;

  if (accountInfo.isNativeOption === 1) {
    accountInfo.rentExemptReserve = u64.fromBuffer(accountInfo.isNative);
    accountInfo.isNative = true;
  } else {
    accountInfo.rentExemptReserve = null;
    accountInfo.isNative = false;
  }

  if (accountInfo.closeAuthorityOption === 0) {
    accountInfo.closeAuthority = null;
  } else {
    accountInfo.closeAuthority = new PublicKey(accountInfo.closeAuthority);
  }

  return accountInfo;
};

// TODO: expose in spl package
const deserializeMint = (data: Buffer) => {
  if (data.length !== MintLayout.span) {
    throw new Error('Not a valid Mint');
  }

  const mintInfo = MintLayout.decode(data);

  if (mintInfo.mintAuthorityOption === 0) {
    mintInfo.mintAuthority = null;
  } else {
    mintInfo.mintAuthority = new PublicKey(mintInfo.mintAuthority);
  }

  mintInfo.supply = u64.fromBuffer(mintInfo.supply);
  mintInfo.isInitialized = mintInfo.isInitialized !== 0;

  if (mintInfo.freezeAuthorityOption === 0) {
    mintInfo.freezeAuthority = null;
  } else {
    mintInfo.freezeAuthority = new PublicKey(mintInfo.freezeAuthority);
  }

  return mintInfo as MintInfo;
};
